import typing
from commonfate_provider import provider
from dataclasses import dataclass
from typing import Dict, Protocol
from inspect import signature
import abc

from commonfate_provider.dataclass import ModelMeta


class DataClass(Protocol):
    # https://stackoverflow.com/questions/54668000/type-hint-for-an-instance-of-a-non-specific-dataclass
    __dataclass_fields__: Dict


DC = typing.TypeVar("DC", bound=DataClass)

LoaderFunc = typing.Callable[[typing.Any], None]


_RESOURCE_LOADERS: typing.Dict[str, LoaderFunc] = {}

P = typing.TypeVar("P", bound=provider.Provider)


class Task(metaclass=ModelMeta):
    def __init__(self, **data: typing.Any) -> None:
        setattr(self, "__dict__", data)

    def json(self) -> dict:
        return {"name": self.__class__.__name__, "ctx": self.__dict__}

    def run(self, p) -> None:
        """
        Runs the task to fetch resources.
        """
        raise Exception("the run method must be implemented")


_PENDING_TASKS: typing.List[Task] = []


def _reset():
    global _PENDING_TASKS
    _PENDING_TASKS = []


def call(task: Task):
    """
    Registers the intent to call an async task. The task will be deferred
    and executed in the future.
    """
    _PENDING_TASKS.append(task)


def _execute(provider: provider.Provider, name: str, ctx: dict):
    """
    Actually execute a task.

    The task can either be a top-level resource loader, defined with the
    `@resources.fetcher` decorator, or a subtask class
    (e.g. `class MyTask(tasks.Task)`).

    Top-level resource loaders do not accept any context values.
    """
    # check if we have a top-level resource loader registered under the name
    resource_loader = _RESOURCE_LOADERS.get(name)
    if resource_loader is not None:
        return resource_loader(provider)

    for Klass in Task.__subclasses__():
        # todo: handle ambiguity in task class naming
        if Klass.__name__ == name:
            task = Klass(**ctx)
            return task.run(provider)

    # if we get here, we couldn't find the task.
    raise Exception(f"could not find task {name}")


def get() -> typing.List[Task]:
    return _PENDING_TASKS



from commonfate_provider import provider, args, resources, tasks
import typing

from pydantic import BaseModel, Field


class GrantData(BaseModel):
    subject: str
    args: typing.Dict[str, str]


class Grant(BaseModel):
    type: typing.Literal["grant"]
    data: GrantData


class Revoke(BaseModel):
    type: typing.Literal["revoke"]
    data: GrantData


class Schema(BaseModel):
    type: typing.Literal["schema"]


class Describe(BaseModel):
    type: typing.Literal["describe"]


class Options(BaseModel):
    class Data(BaseModel):
        arg: str
        """the argument to fetch options for"""

    type: typing.Literal["options"]
    data: Data


class LoadResources(BaseModel):
    class Data(BaseModel):
        name: str
        """the resource loader function ID to run"""
        ctx: dict = {}
        """context information for the task"""

    type: typing.Literal["loadResources"]
    data: Data


class Event(BaseModel):
    __root__: typing.Union[
        Grant, Revoke, Options, LoadResources, Schema, Describe
    ] = Field(..., discriminator="type")


class AWSLambdaRuntime:
    def __init__(
        self, provider: provider.Provider, args_cls: typing.Type[args.Args]
    ) -> None:
        self.provider = provider
        self.args_cls = args_cls

    def handle(self, event, context):
        parsed = Event.parse_obj(event)
        event = parsed.__root__

        if isinstance(event, Grant):
            args = self.args_cls(event.data.args)
            grant = provider._get_grant_func()
            grant(self.provider, event.data.subject, args)
            return {"message": "granting access"}

        elif isinstance(event, Revoke):
            args = self.args_cls(event.data.args)
            revoke = provider._get_revoke_func()
            revoke(self.provider, event.data.subject, args)
            return {"message": "revoking access"}

        if isinstance(event, Options):
            self.args_cls.options(self.provider, event.data.arg)

        if isinstance(event, Schema):
            print("starting to get schema")
            return {"target": self.args_cls.export_schema()}

        if isinstance(event, Describe):
            # Describe returns the configuration of the provider including the current status.
            result = {}
            result["provider"] = self.provider.describe()
            result["config"] = self.provider.config_dict
            result["configValidation"] = self.provider.validate_config()
            result["schema"] = {}
            result["schema"]["target"] = self.args_cls.export_schema()
            result["schema"]["audit"] = resources.audit_schema()

            return result

        elif isinstance(event, LoadResources):
            resources._reset()
            tasks._reset()
            tasks._execute(
                provider=self.provider, name=event.data.name, ctx=event.data.ctx
            )
            # find the resources and pending tasks, and return them
            found = resources.get()
            pending_tasks = tasks.get()
            print(found)
            response = {
                "resources": [r.export_json() for r in found],
                "pendingTasks": [t.json() for t in pending_tasks],
            }
            print(response)
            return response

        else:
            raise Exception(f"unhandled event type")



from commonfate_provider import resources


class ExampleResource(resources.Resource):
    value: str


def test_resource_equality():
    first = ExampleResource(id="test", value="test")
    second = ExampleResource(id="test", value="test")
    assert first.__eq__ is not None
    assert first == second


def test_json_storage_works():
    js = resources.JSONStorage(
        resources=[{"type": "ExampleResource", "id": "test", "value": "test"}]
    )

    want = [ExampleResource(id="test", value="test")]
    got = js.all(ExampleResource)
    assert want == got



from dataclasses import dataclass
import typing


@dataclass
class Log:
    level: typing.Literal["info", "error"]
    msg: str


class Logs:
    logs: typing.List[Log]

    def __init__(self) -> None:
        self.logs = []

    def info(self, msg: str):
        self.logs.append(Log("info", msg=msg))

    def error(self, msg: str):
        self.logs.append(Log("error", msg=msg))

    def succeeded(self) -> bool:
        has_error = any([l for l in self.logs if l.level == "error"])
        return not has_error



from dataclasses import dataclass
from enum import Enum
import typing
from abc import ABC, abstractmethod
from commonfate_provider import provider
from commonfate_provider.dataclass import ModelMeta


class ParseError(Exception):
    """Raised if Provider arguments cannot be parsed"""


@dataclass
class Option:
    value: str
    label: str
    description: typing.Optional[str] = None


@dataclass
class GroupOption:
    children: typing.List[str]
    value: str
    label: str
    label_prefix: typing.Optional[str] = None
    description: typing.Optional[str] = None


class Args(metaclass=ModelMeta):
    def __init__(self, raw_args: dict) -> None:
        """
        initialise Args based on a provided dictionary.
        """
        all_vars = [k for k in vars(self.__class__).keys() if not k.startswith("__")]
        for k in all_vars:
            if k not in raw_args:
                raise ParseError(f"{k} argument is required")
            val = raw_args[k]
            setattr(self, k, val)

    @classmethod
    def options(cls, provider: provider.Provider, arg: str) -> typing.List[Option]:
        val = getattr(cls, arg)
        if not isinstance(val, Field):
            raise Exception(f"invalid arg: {arg}")
        if val.fetch_options is None:
            raise Exception(f"argument {arg} does not provide options")
        return val.fetch_options(provider)

    @classmethod
    def export_schema(cls) -> dict:
        """
        Exports the argument schema defined in an Args class
        to a dictionary.

        For example, if Args is defined as:
        ```
        class Args(args.Args):
            group = args.String(title="Group", description="group to grant access to")
        ```

        this method will produce a dictionary in the form:
        ```json
        {
            "group": {
                "id": "group",
                "description": "group to grant access to",
                "title": "Group",
                "type": "string"
            }
        }
        ```
        """
        arg_schema = {}
        all_vars = [(k, v) for (k, v) in vars(cls).items() if not k.startswith("__")]
        for k, v in all_vars:
            if type(v) == String:
                val: String = v
                schema = {
                    "id": k,
                    "type": "string",
                    "title": val.title,
                    "options": val.fetch_options is not None,
                    "groups": None,
                    "ruleFormElement": FormElement.INPUT,
                    "resourceName": None,
                }
                if val.description is not None:
                    schema["description"] = val.description
                if val.request_element is not None:
                    schema["requestFormElement"] = val.request_element
                if val.rule_element is not None:
                    schema["ruleFormElement"] = val.rule_element

                if val.groups is not None:
                    group_schema = {}
                    for g in val.groups:
                        group_schema[g.__name__] = g.__dict__

                arg_schema[k] = schema
            if type(v) == Resource:
                val: Resource = v
                schema = {
                    "id": k,
                    "type": "string",
                    "title": val.title,
                    "options": val.fetch_options is not None,
                    "groups": None,
                    "ruleFormElement": FormElement.INPUT,
                    "resourceName": val.resource.__name__,
                }
                if val.description is not None:
                    schema["description"] = val.description
                if val.request_element is not None:
                    schema["requestFormElement"] = val.request_element
                if val.rule_element is not None:
                    schema["ruleFormElement"] = val.rule_element

                if val.groups is not None:
                    group_schema = {}
                    for g in val.groups:
                        group_schema[g.__name__] = g.__dict__

                arg_schema[k] = schema

        return arg_schema


class FormElement(str, Enum):
    SELECT = "SELECT"
    INPUT = "INPUT"
    MULTISELECT = "MULTISELECT"


FormElementValue = typing.Literal["SELECT", "INPUT", "MULTISELECT"]


@dataclass
class Group(ABC):
    description: typing.Optional[str] = None
    title: typing.Optional[str] = None
    resource: typing.Optional[typing.Any] = None

    @abstractmethod
    def match(self, key: str) -> typing.List[str]:
        pass


OptionsFetcher = typing.Callable[[typing.Any], typing.List[Option]]


@dataclass
class Field:
    title: typing.Optional[str] = None
    allow_multiple: bool = True
    description: typing.Optional[str] = None
    groups: typing.Optional[typing.Tuple[typing.Type[Group], ...]] = None
    fetch_options: typing.Optional[OptionsFetcher] = None
    request_element: typing.Optional[FormElementValue] = None
    rule_element: typing.Optional[FormElementValue] = None


class String(Field):
    pass


@dataclass
class Resource(Field):
    resource: typing.Optional[typing.Any] = None


@dataclass
class Output:
    resource: typing.Optional[typing.Any] = None



import typing
import pytest
from commonfate_provider import args
from commonfate_provider import provider


def fetch_groups(_) -> typing.List[args.Option]:
    return [
        args.Option(value="one", label="one"),
        args.Option(value="two", label="two"),
        args.Option(value="three", label="three"),
    ]


class ExampleArgs(args.Args):
    group = args.String(fetch_options=fetch_groups)


class BasicProvider(provider.Provider):
    pass


def test_parse_args():
    got_args = ExampleArgs({"group": "test"})
    assert got_args.group == "test"


def test_parse_args_missing_required():
    with pytest.raises(args.ParseError):
        ExampleArgs({})


class ArgsWithOptions(args.Args):
    group = args.String(fetch_options=lambda p: [])


def test_export_schema_with_options_works():
    got = ArgsWithOptions.export_schema()
    assert got["group"]["options"] == True


def test_arg_options_works():
    p = BasicProvider(config_loader=provider.NoopLoader())
    got = ExampleArgs.options(p, "group")
    # should get the 3 options in fetch_groups
    assert len(got) == 3
